/*
    Copyright 2010 Udo Richter.

    Contains code from Klaus Schmidinger's VDR,
    and concepts from Marten Richter

    naludump is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    naludump is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with naludump; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "remux.h"
#include "tools.h"
#include "ringbuffer.h"


long long int TSPacketNumber = 0;
long long int DroppedPackets = 0;

cRingBufferLinear inBuffer(65536, TS_SIZE);
cRingBufferLinear outBuffer(65536);
cNaluDumper NaluDumper;
cPatPmtParser PatPmtParser;

int main(int argc, const char* argv[])
{
    if (argc < 2) {
        printf("Too few arguments");
        exit(1);
    }
    printf("Input file: %s\n", argv[1]);
    int in = open(argv[1], O_RDONLY|O_LARGEFILE);
    if (in == -1) {
        printf("Cannot open %s", argv[1]);
        exit(2);
    }
    int out = 0;
    if (argv[2]) {
        printf("Output file: %s\n", argv[2]);
        out=open(argv[2], O_WRONLY|O_CREAT|O_TRUNC|O_LARGEFILE, DEFFILEMODE);
        if (out == -1) {
            printf("Cannot create %s", argv[2]);
            exit(3);
        }
    }

    while (true) {
        int Length = 0;
        uchar *Data = inBuffer.Get(Length);
        if (Length > 0 && Data[0] != TS_SYNC_BYTE) {
             int Skipped = 1;
             while (Skipped < Length && (Data[Skipped] != TS_SYNC_BYTE || (Length - Skipped > TS_SIZE && Data[Skipped + TS_SIZE] != TS_SYNC_BYTE)))
                Skipped++;
             esyslog("ERROR: skipped %d bytes to sync on start of TS packet", Skipped);
             inBuffer.Del(Skipped);
             continue;
        }

        if (Length < TS_SIZE) { // need more data
            int Count = inBuffer.Read(in);
            if (Count > 0)
                continue; // new data
            if (Count == 0)
                break; // eof
            if (errno == EAGAIN)
                continue;
            printf("Error reading file.\n");
            exit(4);
        }

        bool Drop = false;

        int Pid = TsPid(Data);
        if (Pid == 0)
            PatPmtParser.ParsePat(Data, TS_SIZE);
        else if (Pid == PatPmtParser.PmtPid())
            PatPmtParser.ParsePmt(Data, TS_SIZE);
        else if (Pid == 0x1fff)
            Drop = true;
        else if (Pid == PatPmtParser.Vpid() && PatPmtParser.Vtype() == 0x1B)
            Drop = NaluDumper.ProcessTSPacket(Data);

        if (outBuffer.Free() < TS_SIZE)
        {
            // Flush some data
            while (true)
            {
                int Length = 0;
                uchar *Data = outBuffer.Get(Length);
                Length = Length & ~0xfff; // chunks of 4k
                if (Length==0)
                    break;
                if (out)
                    write(out, Data, Length);
                outBuffer.Del(Length);
            }
        }

        if (!Drop)
        {
            int bytes = outBuffer.Put(Data, TS_SIZE);
            if (bytes != TS_SIZE)
            {
                printf("Fatal internal error: Buffer full\n");
                exit(5);
            }
        }

        inBuffer.Del(TS_SIZE);

        TSPacketNumber++;
        if (Drop)
            DroppedPackets++;
        if (TSPacketNumber%10000==0) {
            printf("\rPackets: %lli Dropped: %lli (%lli%%)", TSPacketNumber, DroppedPackets, TSPacketNumber ? DroppedPackets*100/TSPacketNumber : 0);
            fflush(stdout);
        }
    }

    // Flush out buffer
    while (true)
    {
        int Length = 0;
        uchar *Data = outBuffer.Get(Length);
        if (Length == 0)
            break;
        if (out)
            write(out, Data, Length);
        outBuffer.Del(Length);
    }

    printf("\rPackets: %lli Dropped: %lli (%lli%%)\n", TSPacketNumber, DroppedPackets, TSPacketNumber ? DroppedPackets*100/TSPacketNumber : 0);
    close(in);
    if (out)
        close(out);
}




